/// Events used by applications interacting with the SSH connection
///
/// These include hostkeys, authentication, and shell/command sessions

use self::channel::Channel;

#[allow(unused_imports)]
use {
    crate::error::{Error, Result, TrapBug},
    log::{debug, error, info, log, trace, warn},
};

use core::mem::Discriminant;
use core::fmt::Debug;

use crate::*;
use sshwire::TextString;
use packets::Packet;
use channel::CliSessionOpener;

#[derive(Debug)]
pub enum Event<'g, 'a> {
    Cli(CliEvent<'g, 'a>),
    Serv(ServEvent<'g, 'a>),
    Defunct,
    /// Connection state has progressed
    ///
    /// Should poll `Runner::progress()` again
    Progressed,
    /// No event
    ///
    /// No progress, may idle waiting for external events.
    None,
}

impl<'g, 'a> Event<'g, 'a> {
    pub(crate) fn from_dispatch(disp: DispatchEvent, runner: &'g mut Runner<'a>) -> Result<Self> {
        match disp {
            DispatchEvent::CliEvent(x) => Ok(Self::Cli(x.event(runner)?)),
            DispatchEvent::ServEvent(x) => Ok(Self::Serv(x.event(runner)?)),
            DispatchEvent::Defunct => Ok(Self::Defunct),
            DispatchEvent::None => Ok(Self::None),
            | DispatchEvent::Progressed  => Ok(Self::Progressed),
            // Events handled internally by Runner::progress()
            | DispatchEvent::Data(_)
            => Err(Error::bug()),
        }
    }
}


pub enum CliEvent<'g, 'a>
{
    // AuthRequest { username: TextString<'a>, AuthKind }
    // Banner { banner: TextString<'a>, language: TextString<'a> },
    Hostkey(CheckHostkey<'g, 'a>),
    Username(RequestUsername<'g, 'a>),
    Password(RequestPassword<'g, 'a>),
    Authenticated,
    SessionOpened(CliSessionOpener<'g, 'a>),
    Defunct,

    // Hostkey { /* todo */ },
}

impl Debug for CliEvent<'_, '_> {
    fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
        let e = match self {
            Self::Hostkey(_) => "Hostkey",
            Self::Username(_) => "Username",
            Self::Password(_) => "Password",
            Self::Authenticated => "Authenticated",
            Self::SessionOpened(_) => "SessionOpened",
            Self::Defunct => "Defunct",
        };
        write!(f, "CliEvent({e})")
    }
}

pub struct RequestUsername<'g, 'a> {
    runner: &'g mut Runner<'a>,
}

impl RequestUsername<'_, '_> {
    pub fn respond(self, username: impl AsRef<str>) -> Result<()> {
        self.runner.resume_cliusername(username.as_ref())
    }
}

pub struct RequestPassword<'g, 'a> {
    runner: &'g mut Runner<'a>,
}

impl RequestPassword<'_, '_> {
    pub fn password(self, password: impl AsRef<str>) -> Result<()> {
        self.runner.resume_clipassword(password.as_ref())
    }
}

pub struct CheckHostkey<'g, 'a> {
    runner: &'g mut Runner<'a>,
}

impl CheckHostkey<'_, '_> {
    pub fn hostkey(&self) -> Result<PubKey> {
        self.runner.fetch_checkhostkey()
    }

    pub fn accept(self) -> Result<()> {
        self.runner.resume_checkhostkey(true)
    }

    pub fn reject(self) -> Result<()> {
        self.runner.resume_checkhostkey(false)
    }
}

#[derive(Debug, Clone, Copy)]
pub enum CliEventId {
    Hostkey,
    Username,
    Password,
    Authenticated,
    SessionOpened(ChanNum),
    Defunct

    // TODO:
    // Disconnected
    // Banner,
    // AuthPubkey
    // AgentSign
    // OpenTCPForwarded (new session)
    // TCPDirectOpened (response)
}

impl CliEventId {
    pub fn event<'g, 'a>(self, runner: &'g mut Runner<'a>) -> Result<CliEvent<'g, 'a>> {
        let pk = runner.packet()?;

        match self {
            Self::Username => {
                Ok(CliEvent::Username(RequestUsername { runner }))
            }
            Self::Password => {
                Ok(CliEvent::Password(RequestPassword { runner }))
            }
            Self::Hostkey => {
                debug_assert!(matches!(pk, Some(Packet::KexDHReply(_))));
                Ok(CliEvent::Hostkey(CheckHostkey { runner }))
            }
            Self::Authenticated => Ok(CliEvent::Authenticated),
            Self::SessionOpened(h) => {
                Ok(CliEvent::SessionOpened(runner.cli_session_opener(h)?))
            }
            // (Self::Banner, Packet::UserauthBanner(p)) => {
            //     CliEvent::Banner { banner: p.message, language: p.lang }
            // }
            Self::Defunct => error::BadUsage.fail()
        }
    }

    pub(crate) fn needs_response(&self) -> bool {
        match self {
            | Self::Authenticated
            | Self::SessionOpened(_)
            | Self::Defunct
            => false,
            | Self::Hostkey
            | Self::Username
            | Self::Password
            => true,
        }
    }
}

pub enum ServEvent<'g, 'a> {
    Hostkeys(ServHostkeys<'g, 'a>),
    AuthMethods(ServAuthMethods<'g, 'a>),
    PasswordAuth(ServPasswordAuth<'g, 'a>),
    PubkeyAuth(ServPubkeyAuth<'g, 'a>),
    Defunct,
}

impl Debug for ServEvent<'_, '_> {
    fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
        let e = match self {
            Self::Hostkeys(_) => "Hostkeys",
            Self::AuthMethods(_) => "AuthMethods",
            Self::PasswordAuth(_) => "PasswordAuth",
            Self::PubkeyAuth(_) => "PubkeyAuth",
            Self::Defunct => "Defunct",
        };
        write!(f, "ServEvent({e})")
    }
}

pub struct ServHostkeys<'g, 'a> {
    runner: &'g mut Runner<'a>,
}

impl<'g, 'a> ServHostkeys<'g, 'a> {
    pub fn hostkeys(self, keys: &[&SignKey]) -> Result<()> {
        self.runner.resume_servhostkeys(keys)
    }
}

pub struct ServAuthMethods<'g, 'a> {
    runner: &'g mut Runner<'a>,
    pubkey: bool,
    password: bool,
    done: bool,
}

impl<'a, 'g> ServAuthMethods<'g, 'a> {
    fn new(runner: &'g mut Runner<'a>) -> Self {
        Self {
            runner,
            password: false,
            pubkey: false,
            done: false,
        }
    }

    pub fn username(&self) -> Result<&str> {
        self.raw_username()?.as_str()
    }

    /// Indicate that a client may try password authentication
    pub fn password(&mut self) -> &mut Self {
        self.password = true;
        self
    }

    /// Indicate that a client may try public key authentication
    pub fn pubkey(&mut self) -> &mut Self {
        self.pubkey = true;
        self
    }

    /// Allow log in immediately without any authentication
    pub fn allow_immediately(mut self) -> Result<()> {
        self.done = true;
        self.runner.resume_servauth(true)?;
        todo!()
    }

    pub fn raw_username(&self) -> Result<TextString> {
        self.runner.fetch_servusername()
    }
}

pub struct ServPasswordAuth<'g, 'a> {
    runner: &'g mut Runner<'a>,
    done: bool,
}

impl<'g, 'a> ServPasswordAuth<'g, 'a> {
    fn new(runner: &'g mut Runner<'a>) -> Self {
        Self {
            runner,
            done: false,
        }
    }
    pub fn username(&self) -> Result<&str> {
        self.raw_username()?.as_str()
    }

    pub fn password(&self) -> Result<&str> {
        self.raw_password()?.as_str()
    }

    pub fn allow(mut self) -> Result<()> {
        self.done = true;
        self.runner.resume_servauth(true)
    }

    /// Does not need to be called explicitly, also occurs on drop without `allow()`
    pub fn deny(mut self) -> Result<()> {
        self.done = true;
        self.runner.resume_servauth(false)
    }

    pub fn raw_username(&self) -> Result<TextString> {
        self.runner.fetch_servusername()
    }

    pub fn raw_password(&self) -> Result<TextString> {
        self.runner.fetch_servpassword()
    }
}

// implement Drop to be the same as .deny()
impl Drop for ServPasswordAuth<'_, '_> {
    fn drop(&mut self) {
        if !self.done {
            if let Err(e) = self.runner.resume_servauth(false) {
                trace!("Error for pw auth: {e}")
            }
        }
    }
}

pub struct ServPubkeyAuth<'g, 'a> {
    runner: &'g mut Runner<'a>,
}

impl ServPubkeyAuth<'_, '_> {
    pub fn username(&self) -> Result<&str> {
        self.raw_username()?.as_str()
    }

    pub fn raw_username(&self) -> Result<TextString> {
        self.runner.fetch_servusername()
    }
}


#[derive(Debug, Clone, Copy)]
pub enum ServEventId {
    Hostkeys,
    AuthMethods,
    PasswordAuth,
    PubkeyAuth,
    Defunct,

    // TODO:
    // AuthRequest
    // Disconnected
    // OpenSession
    // OpenTCPForwarded (new session)
    // TCPDirectOpened (response)
    // Banner
    // SessionPty
    // SessionShell
    // SessionExec
}

impl ServEventId {
    pub fn event<'g, 'a>(self, runner: &'g mut Runner<'a>) -> Result<ServEvent<'g, 'a>> {
        let pk = runner.packet()?;

        match self {
            Self::Hostkeys => {
                debug_assert!(matches!(pk, Some(Packet::KexDHInit(_))));
                Ok(ServEvent::Hostkeys(ServHostkeys { runner }))
            }
            Self::AuthMethods => {
                debug_assert!(matches!(pk, Some(Packet::UserauthRequest(_))));
                Ok(ServEvent::AuthMethods(ServAuthMethods::new(runner)))
            }
            Self::PasswordAuth => {
                debug_assert!(matches!(pk, Some(Packet::UserauthRequest(_))));
                Ok(ServEvent::PasswordAuth(ServPasswordAuth::new(runner)))
            }
            Self::PubkeyAuth => {
                debug_assert!(matches!(pk, Some(Packet::UserauthRequest(_))));
                Ok(ServEvent::PubkeyAuth(ServPubkeyAuth { runner }))
            }
            Self::Defunct => Ok(ServEvent::Defunct),
        }
    }

    pub(crate) fn needs_response(&self) -> bool {
        match self {
            | Self::Defunct
            => false,
            | Self::Hostkeys
            | Self::AuthMethods
            | Self::PasswordAuth
            | Self::PubkeyAuth
            => true,
        }
    }
}
