use self::{channel::CliSessionOpener, event::Event};

#[allow(unused_imports)]
use {
    crate::error::{Error, Result, TrapBug},
    log::{debug, error, info, log, trace, warn},
};

use core::{hash::Hash, mem::discriminant, task::{Poll, Waker}};

use pretty_hex::PrettyHex;

use crate::{packets::{Packet, Subsystem}, *};
use packets::{ChannelDataExt, ChannelData};
use crate::channel::{ChanNum, ChanData};
use encrypt::KeyState;
use traffic::{TrafIn, TrafOut};
use event::{CliEvent, ServEvent, ServEventId, CliEventId};

use conn::{Conn, Dispatched, DispatchEvent};

// Runner public methods take a `ChanHandle` which cannot be cloned. This prevents
// confusion if an application were to continue using a channel after the channel
// was completed. The `ChanHandle` is consumed by `Runner::channel_done()`.
// Internally sunset uses `ChanNum`, which is just a newtype around u32.

/// A SSH session instance
///
/// An application provides network or channel data to `Runner` method calls,
/// and provides customisation callbacks via `CliBehaviour` or `ServBehaviour`.
pub struct Runner<'a> {
    conn: Conn,

    /// Binary packet handling from the network buffer
    traf_in: TrafIn<'a>,
    /// Binary packet handling to the network buffer
    traf_out: TrafOut<'a>,

    /// Current encryption/integrity keys
    keys: KeyState,

    /// Waker when output is ready
    output_waker: Option<Waker>,
    /// Waker when ready to consume input.
    input_waker: Option<Waker>,

    closed: bool,

    resume_event: DispatchEvent,
}

impl core::fmt::Debug for Runner<'_> {
    fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
        f.debug_struct("Runner")
        .field("keys", &self.keys)
        .field("output_waker", &self.output_waker)
        .field("input_waker", &self.input_waker)
        .finish_non_exhaustive()
    }
}

// #[derive(Default, Debug, Clone)]
// pub struct Progress<'g, 'a> {
//     pub event: Event<'g, 'a>,
// }

impl<'a> Runner<'a> {
    /// `inbuf` and `outbuf` must be sized to fit the largest SSH packet allowed.
    pub fn new_client(
        inbuf: &'a mut [u8],
        outbuf: &'a mut [u8],
    ) -> Result<Runner<'a>, Error> {
        Self::new(inbuf, outbuf, true)
    }

    /// `inbuf` and `outbuf` must be sized to fit the largest SSH packet allowed.
    pub fn new_server(
        inbuf: &'a mut [u8],
        outbuf: &'a mut [u8],
    ) -> Result<Runner<'a>, Error> {
        Self::new(inbuf, outbuf, false)
    }

    pub fn new(
        inbuf: &'a mut [u8],
        outbuf: &'a mut [u8],
        is_client: bool,
    ) -> Result<Runner<'a>, Error> {
        let conn = Conn::new(is_client)?;
        let runner = Runner {
            conn,
            traf_in: TrafIn::new(inbuf),
            traf_out: TrafOut::new(outbuf),
            keys: KeyState::new_cleartext(),
            output_waker: None,
            input_waker: None,
            closed: false,
            resume_event: Default::default(),
        };

        Ok(runner)
    }

    pub fn is_client(&self) -> bool {
        self.conn.is_client()
    }

    /// Drives connection progress, handling received payload and queueing
    /// packets to send as required.
    pub fn progress(&mut self) -> Result<Event<'_, 'a>>
    {
        let mut disp = Dispatched::default();

        if self.resume_event.take().is_some() {
            if self.resume_event.needs_response() {
                // Events that need a response would have cleared runner.resume_event in their
                // resume handler.
                debug!("No response provided to {:?} event", self.resume_event);
                return error::BadUsage.fail()
            } else {
                self.traf_in.done_payload(false);
                // Clean up
            }
        }

        let mut s = self.traf_out.sender(&mut self.keys);
        // Handle incoming packets
        if let Some((payload, seq)) = self.traf_in.payload() {
            // prog.progressed = true;
            disp = self.conn.handle_payload(payload, seq, &mut s)?;
            trace!("payload disp {disp:?}");

            match disp.event {
                DispatchEvent::Data(data_in) => {
                    // incoming channel data, we haven't finished with payload
                    trace!("handle_payload chan input {data_in:?}");
                    self.traf_in.set_channel_input(data_in)?;
                    // TODO: do we need an Event::Progressed
                    return Ok(Event::None);
                },
                | DispatchEvent::CliEvent(_) 
                | DispatchEvent::ServEvent(_)
                | DispatchEvent::Progressed
                 => {
                    // will return as an event
                }
                DispatchEvent::None => {
                    // no event to return, packets have been completed
                    trace!("handle_payload done");
                    self.traf_in.done_payload(disp.zeroize_payload);
                },
                // TODO, may get used later?
                DispatchEvent::Defunct => return Err(Error::bug()),
            }
        }

        // If there isn't any pending event for the application, run conn.progress()
        // (which may return other events).
        if disp.event.is_none() {
            disp = self.conn.progress(&mut s)?;
            trace!("prog disp {disp:?}");
            match disp.event {
                | DispatchEvent::CliEvent(_) 
                | DispatchEvent::ServEvent(_)
                | DispatchEvent::None
                | DispatchEvent::Progressed
                => (),
                | DispatchEvent::Data(_)
                | DispatchEvent::Defunct
                => return Err(Error::bug()),
            }
        }

        self.wake();

        self.resume_event = disp.event.clone();

        Event::from_dispatch(disp.event, self)
    }

    pub(crate) fn packet(&self) -> Result<Option<packets::Packet>> {
        if let Some((payload, _seq)) = self.traf_in.payload() {
            let packet = self.conn.packet(payload)?;
            Ok(Some(packet))
        } else {
            Ok(None)
        }
    }

    // Accept bytes from the wire, returning the size consumed
    pub fn input(&mut self, buf: &[u8]) -> Result<usize, Error> {
        if self.closed {
            return error::ChannelEOF.fail()
        }
        self.traf_in.input(
            &mut self.keys,
            &mut self.conn.remote_version,
            buf,
        )
    }

    // Whether [`input()`] is ready
    pub fn is_input_ready(&self) -> bool {
        (self.conn.initial_sent() && self.traf_in.is_input_ready()) || self.closed
    }

    /// Set a waker to be notified when [`input()`] is ready to be called.
    pub fn set_input_waker(&mut self, waker: &Waker) {
        if let Some(ref w) = self.input_waker {
            if w.will_wake(waker) {
                return
            }
        }
        if let Some(w) = self.input_waker.replace(waker.clone()) {
            w.wake()
        }
    }


    /// Write any pending output to the wire, returning the size written
    pub fn output(&mut self, buf: &mut [u8]) -> Result<usize, Error> {
        if self.closed {
            return error::ChannelEOF.fail()
        }
        let r = self.traf_out.output(buf);
        if r > 0 {
            trace!("output() wake");
            self.wake();
        }
        Ok(r)
    }

    // Whether [`output()`] is ready
    pub fn is_output_pending(&self) -> bool {
        self.traf_out.is_output_pending() || self.closed
    }

    /// Set a waker to be notified when [`output()`] will have pending data
    pub fn set_output_waker(&mut self, waker: &Waker) {
        if let Some(ref w) = self.output_waker {
            if w.will_wake(waker) {
                return
            }
        }
        if let Some(w) = self.input_waker.replace(waker.clone()) {
            w.wake()
        }
    }

    pub fn close(&mut self) {
        trace!("runner close");
        self.closed = true;
        if let Some(w) = self.output_waker.take() {
            w.wake()
        }
    }

    // TODO: move somewhere client specific?
    pub fn open_client_session(&mut self) -> Result<ChanHandle> {
        trace!("open_client_session");

        let (chan, p) = self.conn.channels.open(packets::ChannelOpenType::Session)?;
        self.traf_out.send_packet(p, &mut self.keys)?;
        self.wake();
        Ok(ChanHandle(chan))
    }

    /// Send data from this application out the wire.
    ///
    /// Returns `Ok(len)` consumed, `Err(Error::ChannelEof)` on EOF,
    /// or other errors.
    pub fn channel_send(
        &mut self,
        chan: &ChanHandle,
        dt: ChanData,
        buf: &[u8],
    ) -> Result<usize> {
        if self.closed {
            return error::ChannelEOF.fail()
        }

        if buf.is_empty() {
            return Ok(0)
        }

        let len = self.ready_channel_send(chan, dt)?;
        let len = match len {
            Some(l) if l == 0 => return Ok(0),
            Some(l) => l,
            None => return Err(Error::ChannelEOF),
        };

        let len = len.min(buf.len());

        let p = self.conn.channels.send_data(chan.0, dt, &buf[..len])?;
        self.traf_out.send_packet(p, &mut self.keys)?;
        self.wake();
        Ok(len)
    }

    /// Receive data coming from the wire into this application.
    ///
    /// Returns `Ok(len)` received, `Err(Error::ChannelEof)` on EOF,
    /// or other errors. Ok(0) indicates no data available, ie pending.
    /// TODO: EOF is unimplemented
    pub fn channel_input(
        &mut self,
        chan: &ChanHandle,
        dt: ChanData,
        buf: &mut [u8],
    ) -> Result<usize> {
        if self.closed {
            return error::ChannelEOF.fail()
        }

        dt.validate_receive(self.conn.is_client())?;

        if self.is_channel_eof(chan) {
            return error::ChannelEOF.fail()
        }

        let (len, complete) = self.traf_in.channel_input(chan.0, dt, buf);
        if let Some(x) = complete {
            self.finished_input(chan, x)?;
        }
        Ok(len)
    }

    /// Receives input data, either normal or extended.
    pub fn channel_input_either(
        &mut self,
        chan: &ChanHandle,
        buf: &mut [u8],
    ) -> Result<(usize, ChanData)> {
        let (len, complete, dt) = self.traf_in.channel_input_either(chan.0, buf);
        if let Some(x) = complete {
            self.finished_input(chan, x)?;
        }
        Ok((len, dt))
    }


    /// Discards any channel input data pending for `chan`, regardless of whether
    /// normal or extended.
    pub fn discard_channel_input(&mut self, chan: &ChanHandle) -> Result<()> {
        let x = self.traf_in.discard_channel_input(chan.0);
        self.finished_input(chan, x)?;
        Ok(())
    }

    fn finished_input(&mut self, chan: &ChanHandle, len: usize) -> Result<()> {
        let mut s = self.traf_out.sender(&mut self.keys);
        self.conn.channels.finished_input(chan.0, len, &mut s)?;
        self.wake();
        Ok(())
    }

    /// Indicates when channel data is ready.
    ///
    /// When channel data is ready, returns a tuple
    /// `Some((channel, data, len))`
    /// `len` is the amount of data ready remaining to read, will always be non-zero.
    /// Note that this returns a `ChanNum` index rather than a `ChanHandle` (which would
    /// be owned by the caller already.
    ///
    /// Returns `None` if no data ready.
    pub fn ready_channel_input(&self) -> Option<(ChanNum, ChanData, usize)> {
        self.traf_in.ready_channel_input()
    }

    pub fn is_channel_eof(&self, chan: &ChanHandle) -> bool {
        self.conn.channels.have_recv_eof(chan.0) || self.closed
    }

    pub fn is_channel_closed(&self, chan: &ChanHandle) -> bool {
        self.conn.channels.is_closed(chan.0) || self.closed
    }

    /// Returns the maximum data that may be sent to a channel
    ///
    /// Returns `Ok(None)` on channel closed.
    ///
    /// May fail with `BadChannelData` if dt is invalid for this session.
    pub fn ready_channel_send(&self, chan: &ChanHandle, dt: ChanData) -> Result<Option<usize>> {
        if self.closed {
            return Ok(None)
        }
        // TODO: return 0 if InKex means we can't transmit packets.

        // Avoid apps polling forever on a packet type that won't come
        dt.validate_send(self.conn.is_client())?;

        // minimum of buffer space and channel window available
        let payload_space = self.traf_out.send_allowed(&self.keys);
        // subtract space for packet headers prior to data
        let payload_space = payload_space.saturating_sub(dt.packet_offset());
        let r = Ok(self.conn.channels.send_allowed(chan.0).map(|s| s.min(payload_space)));
        trace!("ready_channel_send {chan:?} -> {r:?}");
        r
    }

    /// Returns `true` if the channel and `dt` are currently valid for writing.
    ///
    /// Note that they may not be ready to send output.
    pub fn valid_channel_send(&self, chan: &ChanHandle, dt: ChanData) -> bool {
        self.conn.channels.valid_send(chan.0, dt)
    }

    /// Must be called when an application has finished with a channel.
    ///
    /// Channel numbers will not be re-used without calling this, so
    /// failing to call this may result in running out of channels.
    pub fn channel_done(&mut self, chan: ChanHandle) -> Result<()> {
        self.conn.channels.done(chan.0)
    }

    /// Send a terminal window size change report.
    ///
    /// Only call on a client session with a pty
    pub fn term_window_change(&mut self, chan: &ChanHandle, winch: packets::WinChange) -> Result<()> {
        if self.is_client() {
            let mut s = self.traf_out.sender(&mut self.keys);
            self.conn.channels.term_window_change(chan.0, winch, &mut s)
        } else {
            error::BadChannelData.fail()
        }
    }

    pub(crate) fn cli_session_opener(&mut self, ch: ChanNum) -> Result<CliSessionOpener<'_, 'a>> {
        let ch = self.conn.channels.get(ch)?;
        let s = self.traf_out.sender(&mut self.keys);

        Ok(CliSessionOpener {
            ch,
            s,
        })
    }

    fn wake(&mut self) {
        if self.is_input_ready() {
            trace!("wake ready_input, waker {:?}", self.input_waker);
            if let Some(w) = self.input_waker.take() {
                trace!("wake input waker");
                w.wake()
            }
        }

        if self.is_output_pending() {
            if let Some(w) = self.output_waker.take() {
                trace!("wake output waker");
                w.wake()
            } else {
                trace!("no waker");
            }
        }
    }


    fn check_resume_inner(&self, expect: &DispatchEvent,
        compare: &DispatchEvent) {
        match (expect, compare) {
            (DispatchEvent::CliEvent(e), DispatchEvent::CliEvent(c)) => 
                debug_assert_eq!(discriminant(c), discriminant(e),
                    "Expected response to pending {expect:?} event"),
            (DispatchEvent::ServEvent(e), DispatchEvent::ServEvent(c)) => 
                debug_assert_eq!(discriminant(c), discriminant(e),
                    "Expected response to pending {expect:?} event"),
            _ => debug_assert!(false),
        }
    }

    fn resume(&mut self, expect: &DispatchEvent) {
        let prev_event = self.resume_event.take();
        self.check_resume_inner(expect, &prev_event)
    }

    fn check_resume(&self, expect: &DispatchEvent) {
        self.check_resume_inner(expect, &self.resume_event)
    }

    pub(crate) fn resume_cliusername(&mut self, username: &str) -> Result<()> {
        self.resume(&DispatchEvent::CliEvent(CliEventId::Username));
        let mut s = self.traf_out.sender(&mut self.keys);
        self.conn.resume_username(&mut s, username)?;
        self.traf_in.done_payload(false);
        Ok(())
    }

    pub(crate) fn resume_clipassword(&mut self, password: &str) -> Result<()> {
        self.resume(&DispatchEvent::CliEvent(CliEventId::Password));
        let mut s = self.traf_out.sender(&mut self.keys);
        self.conn.resume_password(&mut s, password)?;
        self.traf_in.done_payload(false);
        Ok(())
    }

    pub(crate) fn resume_checkhostkey(&mut self, accept: bool) -> Result<()> {
        self.resume(&DispatchEvent::CliEvent(CliEventId::Hostkey));

        let (payload, _seq) = self.traf_in.payload().trap()?;
        let mut s = self.traf_out.sender(&mut self.keys);

        self.conn.resume_checkhostkey(payload, &mut s, accept)?;
        self.traf_in.done_payload(false);
        Ok(())
    }

    pub(crate) fn fetch_checkhostkey(&self) -> Result<PubKey<'_>> {
        self.check_resume(&DispatchEvent::CliEvent(CliEventId::Hostkey));

        let (payload, _seq) = self.traf_in.payload().trap()?;

        self.conn.fetch_checkhostkey(payload)
    }

    pub(crate) fn resume_servhostkeys(&mut self, keys: &[&SignKey]) -> Result<()> {
        self.resume(&DispatchEvent::ServEvent(ServEventId::Hostkeys));
        let (payload, _seq) = self.traf_in.payload().trap()?;
        let mut s = self.traf_out.sender(&mut self.keys);
        self.conn.resume_servhostkeys(payload, &mut s, keys)?;
        self.traf_in.done_payload(false);
        Ok(())
    }

    pub(crate) fn fetch_servusername(&self) -> Result<TextString> {
        let (payload, _seq) = self.traf_in.payload().trap()?;
        self.conn.fetch_servusername(payload)
    }

    pub(crate) fn fetch_servpassword(&self) -> Result<TextString> {
        let (payload, _seq) = self.traf_in.payload().trap()?;
        self.conn.fetch_servpassword(payload)
    }

    pub(crate) fn resume_servauthmethods(&mut self, have_password: bool, have_pubkey: bool) -> Result<()> {
        todo!()
    }

    pub(crate) fn resume_servauth(&mut self, allow: bool) -> Result<()> {
        let prev_event = self.resume_event.take();
        debug_assert!(
            matches!(prev_event, DispatchEvent::ServEvent(ServEventId::PasswordAuth))
            | matches!(prev_event, DispatchEvent::ServEvent(ServEventId::PubkeyAuth)));

        let (payload, _seq) = self.traf_in.payload().trap()?;
        let mut s = self.traf_out.sender(&mut self.keys);
        self.conn.resume_servauth(payload, &mut s, allow)
    }
}

/// Represents an open channel, owned by the application.
///
/// Must be released by calling [`Runner::channel_done()`]

// Inner contents are crate-private to ensure that arbitrary
// channel numbers cannot be used after closing/reuse.
//
// This must not be `Clone`
pub struct ChanHandle(pub(crate) ChanNum);

impl ChanHandle {
    /// Returns the channel number
    ///
    /// This can be used by applications as an index.
    /// Channel numbers satisfy
    /// `0 <= num < sunset::config::MAX_CHANNELS`.
    ///
    /// An index may be reused after a call to [`Runner::channel_done()`],
    /// applications must take care not to keep using this `num()` index after
    /// that.
    pub fn num(&self) -> ChanNum {
        self.0
    }
}

impl core::fmt::Debug for ChanHandle {
    fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
        write!(f, "ChanHandle({})", self.num())
    }
}

#[cfg(test)]
mod tests {
    // TODO: test send_allowed() limits
}
